# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import xarray as xr

from typing import Optional, Union
from datetime import timedelta

from sklearn.preprocessing import MinMaxScaler
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.model_selection import TimeSeriesSplit, KFold, cross_validate as c_val
from sklearn.utils import shuffle
from sklearn.exceptions import NotFittedError

from gordo.machine.model.base import GordoBase
from gordo.machine.model import utils as model_utils
from gordo.machine.model.models import KerasAutoEncoder
from gordo.machine.model.anomaly.base import AnomalyDetectorBase


class DiffBasedAnomalyDetector(AnomalyDetectorBase):
    def __init__(
        self,
        base_estimator: BaseEstimator = KerasAutoEncoder(kind="feedforward_hourglass"),
        scaler: TransformerMixin = MinMaxScaler(),
        require_thresholds: bool = True,
        shuffle: bool = False,
        window: Optional[int] = None,
        smoothing_method: Optional[str] = None,
    ):
        """
        Estimator which wraps a ``base_estimator`` and provides a diff error
        based approach to anomaly detection.

        It trains a ``scaler`` to the target **after** training, purely for
        error calculations. The underlying ``base_estimator`` is trained
        with the original, unscaled, ``y``.

        Threshold calculation is based on a rolling statistic of the validation errors
        on the last fold of cross-validation.

        Parameters
        ----------
        base_estimator: sklearn.base.BaseEstimator
            The model to which normal ``.fit``, ``.predict`` methods will be used.
            defaults to py:class:`gordo.machine.model.models.KerasAutoEncoder` with
            ``kind='feedforward_hourglass``
        scaler: sklearn.base.TransformerMixin
            Defaults to ``sklearn.preprocessing.RobustScaler``
            Used for transforming model output and the original ``y`` to calculate
            the difference/error in model output vs expected.
        require_thresholds: bool
            Requires calculating ``thresholds_`` via a call to
            :func:`~DiffBasedAnomalyDetector.cross_validate`. If this is set
            (default True), but :func:`~DiffBasedAnomalyDetector.cross_validate` was not
            called before calling :func:`~DiffBasedAnomalyDetector.anomaly`
            an ``AttributeError`` will be raised.
        shuffle: bool
            Flag to shuffle or not data in ``.fit`` so that the model, if relevant,
            will be trained on a sample of data accross the time range and not just
            the last elements according to model arg ``validation_split``.
        window: int
            Window size for smoothed thresholds
        smoothing_method: str
            Method to be used together with ``window`` to smooth metrics.
            Must be one of: 'smm': simple moving median, 'sma': simple moving average or
            'ewma': exponential weighted moving average.
        """
        self.base_estimator = base_estimator
        self.scaler = scaler
        self.require_thresholds = require_thresholds
        self.shuffle = shuffle
        self.window = window
        self.smoothing_method = smoothing_method
        if self.window is not None and self.smoothing_method is None:
            self.smoothing_method = "smm"

    def __getattr__(self, item):
        """
        Treat this model as transparent into base_estimator unless
        referring to something owned by this object
        """
        if item in self.__dict__:
            return getattr(self, item)
        else:
            return getattr(self.base_estimator, item)

    def get_metadata(self):
        """
        Generates model metadata.

        Returns
        -------
        dict
        """
        metadata = dict()
        if hasattr(self, "feature_thresholds_"):
            metadata["feature-thresholds"] = self.feature_thresholds_.tolist()
        if hasattr(self, "aggregate_threshold_"):
            metadata["aggregate-threshold"] = self.aggregate_threshold_
        if hasattr(self, "feature_thresholds_per_fold_"):
            metadata[
                "feature-thresholds-per-fold"
            ] = self.feature_thresholds_per_fold_.to_dict()
        if hasattr(self, "aggregate_thresholds_per_fold_"):
            metadata[
                "aggregate-thresholds-per-fold"
            ] = self.aggregate_thresholds_per_fold_
        # Window threshold metadata
        if hasattr(self, "window"):
            metadata["window"] = self.window
        if hasattr(self, "smoothing_method"):
            metadata["smoothing-method"] = self.smoothing_method
        if (
            hasattr(self, "smooth_feature_thresholds_")
            and self.smooth_aggregate_threshold_ is not None
        ):
            metadata[
                "smooth-feature-thresholds"
            ] = self.smooth_feature_thresholds_.tolist()
        if (
            hasattr(self, "smooth_aggregate_threshold_")
            and self.smooth_aggregate_threshold_ is not None
        ):
            metadata["smooth-aggregate-threshold"] = self.smooth_aggregate_threshold_

        if hasattr(self, "smooth_feature_thresholds_per_fold_"):
            metadata[
                "smooth-feature-thresholds-per-fold"
            ] = self.smooth_feature_thresholds_per_fold_.to_dict()
        if hasattr(self, "smooth_aggregate_thresholds_per_fold_"):
            metadata[
                "smooth-aggregate-thresholds-per-fold"
            ] = self.smooth_aggregate_thresholds_per_fold_

        if isinstance(self.base_estimator, GordoBase):
            metadata.update(self.base_estimator.get_metadata())
        else:
            metadata.update(
                {
                    "scaler": str(self.scaler),
                    "base_estimator": str(self.base_estimator),
                    "shuffle": self.shuffle,
                }
            )
        return metadata

    def score(
        self,
        X: Union[np.ndarray, pd.DataFrame],
        y: Union[np.ndarray, pd.DataFrame],
        sample_weight: Optional[np.ndarray] = None,
    ) -> float:
        return self.base_estimator.score(X, y)

    def get_params(self, deep=True):
        """
        Get parameters for this estimator.

        Returns
        -------
        dict
        """
        params = {
            "base_estimator": self.base_estimator,
            "scaler": self.scaler,
            "shuffle": self.shuffle,
        }
        if self.window is not None:
            params["window"] = self.window
            params["smoothing_method"] = self.smoothing_method
        return params

    def fit(self, X: np.ndarray, y: np.ndarray):
        if self.shuffle:
            X_shuff, y_shuff = shuffle(X, y, random_state=0)
            self.base_estimator.fit(X_shuff, y_shuff)
        else:
            self.base_estimator.fit(X, y)

        self.scaler.fit(y)  # Scaler is used for calculating errors in .anomaly()
        return self

    def cross_validate(
        self,
        *,
        X: Union[pd.DataFrame, np.ndarray],
        y: Union[pd.DataFrame, np.ndarray],
        cv=TimeSeriesSplit(n_splits=3),
        **kwargs,
    ):
        """
        Run TimeSeries cross validation on the model, and will update the model's
        threshold values based on the cross validation folds.

        Parameters
        ----------
        X: Union[pd.DataFrame, np.ndarray]
            Input data to the model
        y: Union[pd.DataFrame, np.ndarray]
            Target data
        kwargs: dict
            Any additional kwargs to be passed to
            :func:`sklearn.model_selection.cross_validate`

        Returns
        -------
        dict
        """
        # Depend on having the trained fold models
        kwargs.update(dict(return_estimator=True, cv=cv))

        cv_output = c_val(self, X=X, y=y, **kwargs)

        self.feature_thresholds_per_fold_ = pd.DataFrame()
        self.aggregate_thresholds_per_fold_ = {}
        self.smooth_feature_thresholds_per_fold_ = pd.DataFrame()
        self.smooth_aggregate_thresholds_per_fold_ = {}
        smooth_aggregate_threshold_fold = None
        smooth_tag_thresholds_fold = None

        for i, ((_, test_idxs), split_model) in enumerate(
            zip(kwargs["cv"].split(X, y), cv_output["estimator"])
        ):
            y_pred = split_model.predict(
                X.iloc[test_idxs] if isinstance(X, pd.DataFrame) else X[test_idxs]
            )

            # Adjust y_true for any possible model offset in its prediction
            test_idxs = test_idxs[-len(y_pred) :]
            y_true = y.iloc[test_idxs] if isinstance(y, pd.DataFrame) else y[test_idxs]

            # Model's timestep scaled mse over all features
            scaled_mse = self._scaled_mse_per_timestep(split_model, y_true, y_pred)

            # Absolute error
            mae = self._absolute_error(y_true, y_pred)

            # For the aggregate threshold for the fold model,
            # use the mse of scaled residuals per timestep
            aggregate_threshold_fold = scaled_mse.rolling(6).min().max()
            self.aggregate_thresholds_per_fold_[f"fold-{i}"] = aggregate_threshold_fold

            # Accumulate the rolling mins of diffs into common df
            tag_thresholds_fold = mae.rolling(6).min().max()
            tag_thresholds_fold.name = f"fold-{i}"
            self.feature_thresholds_per_fold_ = self.feature_thresholds_per_fold_.append(
                tag_thresholds_fold
            )

            if self.window is not None:
                # Calculate smoothed thresholds only if len of data >= window
                smooth_aggregate_threshold_fold = (
                    scaled_mse.rolling(self.window).min().max()
                )
                self.smooth_aggregate_thresholds_per_fold_[
                    f"fold-{i}"
                ] = smooth_aggregate_threshold_fold

                smooth_tag_thresholds_fold = mae.rolling(self.window).min().max()
                smooth_tag_thresholds_fold.name = f"fold-{i}"
                self.smooth_feature_thresholds_per_fold_ = self.smooth_feature_thresholds_per_fold_.append(
                    smooth_tag_thresholds_fold
                )

        # Final thresholds are the thresholds from the last cv split/fold
        self.feature_thresholds_ = tag_thresholds_fold

        # For the aggregate also use the thresholds from the last split/fold
        self.aggregate_threshold_ = aggregate_threshold_fold

        # For the smoothed thresholds also use the last fold
        self.smooth_aggregate_threshold_ = smooth_aggregate_threshold_fold
        self.smooth_feature_thresholds_ = smooth_tag_thresholds_fold

        return cv_output

    @staticmethod
    def _scaled_mse_per_timestep(
        model: BaseEstimator,
        y_true: Union[pd.DataFrame, np.ndarray],
        y_pred: Union[pd.DataFrame, np.ndarray],
    ) -> pd.Series:
        """
        Calculate the scaled MSE per timestep/sample
        Parameters
        ----------
        model: BaseEstimator
            Instance of a fitted :class:`~DiffBasedAnomalyDetector`
        y_true: Union[numpy.ndarray, pd.DataFrame]
        y_pred: Union[numpy.ndarray, pd.DataFrame]
        Returns
        -------
        panadas.Series
            The MSE calculated from the scaled y and y predicted.
        """
        try:
            scaled_y_true = model.scaler.transform(y_true)
        except (NotFittedError, ValueError):
            scaled_y_true = model.scaler.fit_transform(y_true)
        scaled_y_pred = model.scaler.transform(y_pred)
        mse_per_time_step = ((scaled_y_pred - scaled_y_true) ** 2).mean(axis=1)
        return pd.Series(mse_per_time_step)

    @staticmethod
    def _absolute_error(
        y_true: Union[pd.DataFrame, np.ndarray], y_pred: Union[pd.DataFrame, np.ndarray]
    ) -> pd.DataFrame:

        return pd.DataFrame(np.abs(y_true - y_pred))

    def _smoothing(self, metric: Union[pd.DataFrame, pd.Series]):
        if self.smoothing_method == "smm":
            return metric.rolling(self.window).median()
        elif self.smoothing_method == "sma":
            return metric.rolling(self.window).mean()
        elif self.smoothing_method == "ewma":
            return metric.ewm(span=self.window).mean()

    def anomaly(
        self,
        X: Union[pd.DataFrame, xr.DataArray],
        y: Union[pd.DataFrame, xr.DataArray],
        frequency: Optional[timedelta] = None,
    ) -> Union[pd.DataFrame, xr.Dataset]:
        """
        Create an anomaly dataframe from the base provided dataframe.

        Parameters
        ----------
        X: pd.DataFrame
            Dataframe representing the data to go into the model.
        y: pd.DataFrame
            Dataframe representing the target output of the model.

        Returns
        -------
        pd.DataFrame
            A superset of the original base dataframe with added anomaly specific
            features
        """

        # Get the model output, falling back to transform if 'predict' doesn't exist
        model_output = (
            self.predict(X) if hasattr(self, "predict") else self.transform(X)
        )

        # Create the basic dataframe with 'model-output' & 'model-input'
        data = model_utils.make_base_dataframe(
            tags=X.columns,
            model_input=getattr(X, "values", X),
            model_output=model_output,
            target_tag_list=y.columns,
            index=getattr(X, "index", None),
            frequency=frequency,
        )

        model_out_scaled = pd.DataFrame(
            self.scaler.transform(data["model-output"]),
            columns=data["model-output"].columns,
            index=data.index,
        )

        # Calculate the absolute scaled tag anomaly
        # Ensure to offset the y to match model out, which could be less if it is a LSTM
        scaled_y = self.scaler.transform(y)
        tag_anomaly_scaled = np.abs(model_out_scaled - scaled_y[-len(data) :, :])
        tag_anomaly_scaled.columns = pd.MultiIndex.from_product(
            (("tag-anomaly-scaled",), tag_anomaly_scaled.columns)
        )
        data = data.join(tag_anomaly_scaled)

        # Calculate scaled total anomaly
        data["total-anomaly-scaled"] = np.square(data["tag-anomaly-scaled"]).mean(
            axis=1
        )

        # Calculate the absolute unscaled tag anomalies
        unscaled_abs_diff = pd.DataFrame(
            data=np.abs(
                data["model-output"].to_numpy() - y.to_numpy()[-len(data) :, :]
            ),
            index=data.index,
            columns=pd.MultiIndex.from_product(
                (("tag-anomaly-unscaled",), y.columns.tolist())
            ),
        )
        data = data.join(unscaled_abs_diff)

        # Calculate the scaled total anomaly
        data["total-anomaly-unscaled"] = np.square(data["tag-anomaly-unscaled"]).mean(
            axis=1
        )

        if self.window is not None and self.smoothing_method is not None:
            # Calculate scaled tag-level smoothed anomaly scores
            smooth_tag_anomaly_scaled = self._smoothing(tag_anomaly_scaled)
            smooth_tag_anomaly_scaled.columns = smooth_tag_anomaly_scaled.columns.set_levels(
                ["smooth-tag-anomaly-scaled"], level=0
            )
            data = data.join(smooth_tag_anomaly_scaled)

            # Calculate scaled smoothed total anomaly score
            data["smooth-total-anomaly-scaled"] = self._smoothing(
                data["total-anomaly-scaled"]
            )

            # Calculate unscaled tag-level smoothed anomaly scores
            smooth_tag_anomaly_unscaled = self._smoothing(unscaled_abs_diff)

            smooth_tag_anomaly_unscaled.columns = smooth_tag_anomaly_unscaled.columns.set_levels(
                ["smooth-tag-anomaly-unscaled"], level=0
            )
            data = data.join(smooth_tag_anomaly_unscaled)

            # Calculate unscaled smoothed total anomaly score
            data["smooth-total-anomaly-unscaled"] = self._smoothing(
                data["total-anomaly-unscaled"]
            )

        # If we have `thresholds_` values, then we can calculate anomaly confidence
        confidence, index = None, None

        if hasattr(self, "feature_thresholds_"):
            confidence = unscaled_abs_diff.values / self.feature_thresholds_.values
            index = unscaled_abs_diff.index

        if confidence is not None and index is not None:
            # Dataframe of % abs_diff is of the thresholds
            # This is now based on the smoothed tag anomaly
            anomaly_confidence_scores = pd.DataFrame(
                confidence,
                index=index,
                columns=pd.MultiIndex.from_product(
                    (("anomaly-confidence",), data["model-output"].columns)
                ),
            )
            data = data.join(anomaly_confidence_scores)

        total_anomaly_confidence = None

        if hasattr(self, "aggregate_threshold_"):
            total_anomaly_confidence = (
                data["total-anomaly-scaled"] / self.aggregate_threshold_
            )

        if total_anomaly_confidence is not None:
            data["total-anomaly-confidence"] = total_anomaly_confidence

        # Explicitly raise error if we were required to do threshold based calculations
        # should would have required a call to .cross_validate before .anomaly
        if self.require_thresholds and not any(
            hasattr(self, attr)
            for attr in ("feature_thresholds_", "aggregate_threshold_")
        ):
            raise AttributeError(
                f"`require_thresholds={self.require_thresholds}` however "
                f"`.cross_validate` needs to be called in order to calculate these"
                f"thresholds before calling `.anomaly`"
            )

        return data


class DiffBasedKFCVAnomalyDetector(DiffBasedAnomalyDetector):
    def __init__(
        self,
        base_estimator: BaseEstimator = KerasAutoEncoder(kind="feedforward_hourglass"),
        scaler: TransformerMixin = MinMaxScaler(),
        require_thresholds: bool = True,
        shuffle: bool = True,
        window: int = 144,
        smoothing_method: str = "smm",
        threshold_percentile: float = 0.99,
    ):
        """
        Estimator which wraps a ``base_estimator`` and provides a diff error
        based approach to anomaly detection.

        It trains a ``scaler`` to the target **after** training, purely for
        error calculations. The underlying ``base_estimator`` is trained
        with the original, unscaled, ``y``.

        Threshold calculation is based on a percentile of the smoothed validation
        errors as calculated from cross-validation predictions.

        Parameters
        ----------
        base_estimator: sklearn.base.BaseEstimator
            The model to which normal ``.fit``, ``.predict`` methods will be used.
            defaults to py:class:`gordo.machine.model.models.KerasAutoEncoder` with
            ``kind='feedforward_hourglass``
        scaler: sklearn.base.TransformerMixin
            Defaults to ``sklearn.preprocessing.RobustScaler``
            Used for transforming model output and the original ``y`` to calculate
            the difference/error in model output vs expected.
        require_thresholds: bool
            Requires calculating ``thresholds_`` via a call to
            :func:`~DiffBasedAnomalyDetector.cross_validate`.
            If this is set (default True), but
            :func:`~DiffBasedAnomalyDetector.cross_validate` was not called before
            calling :func:`~DiffBasedAnomalyDetector.anomaly` an ``AttributeError``
            will be raised.
        shuffle: bool
            Flag to shuffle or not data in ``.fit`` so that the model, if relevant,
            will be trained on a sample of data accross the time range and not just
            the last elements according to model arg ``validation_split``.
        window: int
            Window size for smooth metrics and threshold calculation.
        smoothing_method: str
            Method to be used together with ``window`` to smooth metrics.
            Must be one of: 'smm': simple moving median, 'sma': simple moving average or
            'ewma': exponential weighted moving average.
        threshold_percentile: float
            Percentile of the validation data to be used to calculate the threshold.
        """
        self.base_estimator = base_estimator
        self.scaler = scaler
        self.require_thresholds = require_thresholds
        self.window = window
        self.shuffle = shuffle
        self.smoothing_method = smoothing_method
        self.threshold_percentile = threshold_percentile

    def get_params(self, deep=True):
        """
        Get parameters for this estimator.

        Returns
        -------
        dict
        """
        params = {
            "base_estimator": self.base_estimator,
            "scaler": self.scaler,
            "window": self.window,
            "smoothing_method": self.smoothing_method,
            "shuffle": self.shuffle,
            "threshold_percentile": self.threshold_percentile,
        }
        return params

    def get_metadata(self):
        """
        Generates model metadata.

        Returns
        -------
        dict
        """
        metadata = dict()

        if hasattr(self, "feature_thresholds_"):
            metadata["feature-thresholds"] = self.feature_thresholds_.tolist()
        if hasattr(self, "aggregate_threshold_"):
            metadata["aggregate-threshold"] = self.aggregate_threshold_
        if isinstance(self.base_estimator, GordoBase):
            metadata.update(self.base_estimator.get_metadata())
        else:
            metadata.update(
                {
                    "scaler": str(self.scaler),
                    "base_estimator": str(self.base_estimator),
                    "shuffle": self.shuffle,
                    "window": self.window,
                    "smoothing-method": self.smoothing_method,
                    "threshold-percentile": self.threshold_percentile,
                }
            )
        return metadata

    def cross_validate(
        self,
        *,
        X: Union[pd.DataFrame, np.ndarray],
        y: Union[pd.DataFrame, np.ndarray],
        cv=KFold(n_splits=5, shuffle=True, random_state=0),
        **kwargs,
    ):
        """
        Run Kfold cross validation on the model, and will update the model's threshold
        values based on a percentile of the validation metrics.

        Parameters
        ----------
        X: Union[pd.DataFrame, np.ndarray]
            Input data to the model
        y: Union[pd.DataFrame, np.ndarray]
            Target data
        kwargs: dict
            Any additional kwargs to be passed to
            :func:`sklearn.model_selection.cross_validate`

        Returns
        -------
        dict
        """

        # Depend on having the trained fold models
        kwargs.update(dict(return_estimator=True, cv=cv))

        cv_output = c_val(self, X=X, y=y, **kwargs)

        # Create empty dataframes to hold fold data
        y_pred = pd.DataFrame(
            np.zeros_like(y),
            index=getattr(y, "index", None),
            columns=getattr(y, "columns", None),
        )
        y = pd.DataFrame(y)
        y_val_mse = pd.Series(index=getattr(y, "index", None))

        # Calculate per-fold validation metrics
        for i, ((_, test_idxs), split_model) in enumerate(
            zip(kwargs["cv"].split(X, y), cv_output["estimator"])
        ):
            y_pred.iloc[test_idxs] = split_model.predict(
                X.iloc[test_idxs].to_numpy()
                if isinstance(X, pd.DataFrame)
                else X[test_idxs]
            )

            y_val_mse.iloc[test_idxs] = self._scaled_mse_per_timestep(
                split_model, y.iloc[test_idxs], y_pred.iloc[test_idxs]
            ).to_numpy()

        # Calculate aggregate threshold
        self.aggregate_threshold_ = self._calculate_threshold(y_val_mse)

        # Calculate tag thresholds
        self.feature_thresholds_ = self._calculate_feature_thresholds(y, y_pred)

        return cv_output

    def _calculate_feature_thresholds(
        self, y_true: pd.DataFrame, y_pred: pd.DataFrame
    ) -> np.ndarray:
        absolute_error = self._absolute_error(y_true, y_pred)
        return self._calculate_threshold(absolute_error)

    def _calculate_threshold(
        self, validation_metric: Union[pd.DataFrame, pd.Series]
    ) -> Union[float, pd.Series]:
        val_metric = self._smoothing(validation_metric)
        return val_metric.quantile(self.threshold_percentile)



# -*- coding: utf-8 -*-

import inspect
import logging

from sklearn.pipeline import Pipeline


logger = logging.getLogger(__name__)


def into_definition(pipeline: Pipeline, prune_default_params: bool = False) -> dict:
    """
    Convert an instance of ``sklearn.pipeline.Pipeline`` into a dict definition
    capable of being reconstructed with
    ``gordo.serializer.from_definition``

    Parameters
    ----------
    pipeline: sklearn.pipeline.Pipeline
        Instance of pipeline to decompose
    prune_default_params: bool
        Whether to prune the default parameters found in current instance of the transformers
        vs what their default params are.

    Returns
    -------
    dict
        definitions for the pipeline, compatible to be reconstructed with
        :func:`gordo.serializer.from_definition`

    Example
    -------
    >>> import yaml
    >>> from sklearn.pipeline import Pipeline
    >>> from sklearn.decomposition import PCA
    >>> from gordo.machine.model.models import KerasAutoEncoder
    >>>
    >>> pipe = Pipeline([('pca', PCA(4)), ('ae', KerasAutoEncoder(kind='feedforward_model'))])
    >>> pipe_definition = into_definition(pipe)  # It is now a standard python dict of primitives.
    >>> print(yaml.dump(pipe_definition))
    sklearn.pipeline.Pipeline:
      memory: null
      steps:
      - sklearn.decomposition._pca.PCA:
          copy: true
          iterated_power: auto
          n_components: 4
          random_state: null
          svd_solver: auto
          tol: 0.0
          whiten: false
      - gordo.machine.model.models.KerasAutoEncoder:
          kind: feedforward_model
      verbose: false
    <BLANKLINE>
    """
    steps = _decompose_node(pipeline, prune_default_params)
    return steps


def _decompose_node(step: object, prune_default_params: bool = False):
    """
    Decompose a specific instance of a scikit-learn transformer,
    including Pipelines or FeatureUnions

    Parameters
    ----------
    step
        An instance of a Scikit-Learn transformer class
    prune_default_params
        Whether to output the default parameter values into the definition. If True,
        only those parameters differing from the default params will be output.

    Returns
    -------
    dict
        decomposed node - Where key is the import string for the class and associated value
        is a dict of parameters for that class.
    """

    import_str = f"{step.__module__}.{step.__class__.__name__}"

    if hasattr(step, "into_definition"):
        definition = getattr(step, "into_definition")()
    else:
        params = getattr(step, "get_params")(deep=False)

        definition = load_definition_from_params(params)

        definition = (
            _prune_default_parameters(step, definition)
            if prune_default_params
            else definition
        )
    return {import_str: definition}


def _prune_default_parameters(obj: object, current_params) -> dict:
    """
    Take an instance of an object and determine what the default parameters are
    against what its current parameters are.

    Parameters
    ----------
        obj: object - An instance of an object
        current_params: dict - A mapping of current parameters for the obj

    Returns
    -------
        dict - Containing only parameters which are different from default
    """

    signature = inspect.signature(obj.__class__.__init__)
    default_params = {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }
    logger.debug(f"Current params: {current_params}, default_params: {default_params}")

    return {
        k: v
        for (k, v) in current_params.items()
        if current_params[k] != default_params[k]
    }


def load_definition_from_params(params: dict) -> dict:
    """
    Recursively decomposing each of values from params into the definition

    Parameters
    ----------
        params: dict

    Returns
    -------
        dict

    """
    definition = {}
    for param, param_val in params.items():

        if hasattr(param_val, "get_params") or hasattr(param_val, "into_definition"):
            definition[param] = _decompose_node(param_val)

        # Handle parameter value that is a list
        elif isinstance(param_val, list):

            # Decompose second elements; these are tuples of (str, BaseEstimator)
            # or list of other types such as ints.
            # TODO: Make this more robust, probably via another function to parse the iterable recursively
            # TODO: b/c it _could_, in theory, be a dict of {str: BaseEstimator} or similar.
            definition[param] = [
                _decompose_node(leaf[1]) if isinstance(leaf, tuple) else leaf
                for leaf in param_val
            ]

        # Handle FunctionTransformer function object type parameters
        elif callable(param_val):
            # param_val is a function for FunctionTransformer.func init param
            definition[param] = f"{param_val.__module__}.{param_val.__name__}"

        else:
            definition[param] = param_val
    return definition



# -*- coding: utf-8 -*-
import random
import pytest
import pandas as pd
import numpy as np
import dateutil

from gordo.server import utils as server_utils


@pytest.mark.parametrize(
    "df",
    (
        pd.DataFrame(np.random.random((10, 10))),
        pd.DataFrame(
            np.random.random((10, 10)),
            index=pd.date_range(start="2016-01-01", end="2016-01-02", periods=10),
        ),
        pd.DataFrame(
            np.random.random((10, 4)),
            columns=pd.MultiIndex.from_product((("col1", "col2"), ("ft1", "ft2"))),
        ),
    ),
)
def test_dataframe_parquet_serializers(df):
    """The (de)serialization functions should be interoperable"""
    serialized = server_utils.dataframe_into_parquet_bytes(df.copy())
    df_clone = server_utils.dataframe_from_parquet_bytes(serialized)
    assert df.columns.tolist() == df_clone.columns.tolist()
    assert df.index.tolist() == df_clone.index.tolist()
    assert np.allclose(df.values, df_clone.values)


@pytest.mark.parametrize(
    "df",
    [
        # Multi-column
        pd.DataFrame(
            np.random.random((10, 4)),
            columns=pd.MultiIndex.from_product(
                (("feature1", "feature2"), ("col1", "col2"))
            ),
            index=pd.date_range("2016-01-01", "2016-02-01", periods=10),
        ),
        # Normal dataframe, no date index
        pd.DataFrame(
            np.random.random((10, 4)), columns=["col1", "col2", "col3", "col4"]
        ),
    ],
)
def test_dataframe_from_to_dict(df):
    """
    Test (de)serializations back and forth between dataframe -> dict -> dataframe
    """
    index_was_datetimes: bool = isinstance(df.index, pd.DatetimeIndex)

    cloned = server_utils.dataframe_from_dict(server_utils.dataframe_to_dict(df))

    if index_was_datetimes:
        # Ensure the function hasn't mutated the index.
        assert isinstance(df.index, pd.DatetimeIndex)

    assert np.allclose(df.values, cloned.values)
    assert df.columns.tolist() == cloned.columns.tolist()
    assert df.index.tolist() == cloned.index.tolist()


@pytest.mark.parametrize(
    "expect_multi_lvl, data",
    [
        (False, {"col1": [0, 1, 2, 3], "col2": [0, 1, 2, 3]}),
        (True, {("ft1", "col1"): [0, 1, 2, 3], ("ft1", "col2"): [0, 1, 2, 3]}),
        (True, {"ft1": {"col1": [0, 1, 2]}, "ft2": {"col1": [0, 1, 2]}}),
        (False, [[0, 1, 2], [0, 1, 2]]),
    ],
)
def test_dataframe_to_from_dict(expect_multi_lvl: bool, data: dict):
    """
    Creating dataframes from various raw data structures should have determined behavior
    such as not creating MultiIndex columns with a dict of simple key to array mappings.
    """
    df = server_utils.dataframe_from_dict(data)
    if expect_multi_lvl:
        assert isinstance(df.columns, pd.MultiIndex)
    else:
        assert not isinstance(df.columns, pd.MultiIndex)


@pytest.mark.parametrize(
    "index",
    (
        range(10),
        map(str, range(10)),
        map(str, random.sample(range(10), 10)),
        pd.date_range(start="2020-01-01", end="2020-01-02", periods=10),
        pd.date_range(start="2020-01-01", end="2020-01-02", periods=10).astype(str),
        random.sample(
            pd.date_range(start="2020-01-01", end="2020-01-02", periods=10).tolist(), 10
        ),
        random.sample(
            pd.date_range(start="2020-01-01", end="2020-01-02", periods=10)
            .astype(str)
            .tolist(),
            10,
        ),
    ),
)
def test_dataframe_from_dict_ordering(index):
    """
    We expect that from_dict should order based on the index, and will parse the index
    either as datetime or integers and sort in ascending order from there.
    """
    df = pd.DataFrame(np.random.random((10, 5)))
    df.index = index
    original = df.copy()

    # What we want
    if isinstance(original.index[0], str):
        # Parse as datetime or integers if index is string
        try:
            original.index = original.index.map(dateutil.parser.isoparse)
        except ValueError:
            original.index = original.index.map(int)
    original.sort_index(inplace=True)

    # What we get
    df_out = server_utils.dataframe_from_dict(server_utils.dataframe_to_dict(df))

    assert np.alltrue(df_out.index == original.index)
    assert np.alltrue(df_out.values == original.values)



import pytest
import peewee

from gordo.machine import Machine
from gordo.reporters.postgres import PostgresReporter, Machine as PostgresMachine


def test_builder_with_reporter(postgresdb, metadata):
    """
    Verify a model can take a reporter and .report() will run any given reporters
    """
    reporter = PostgresReporter(host="localhost")
    metadata["runtime"]["reporters"].append(reporter.to_dict())

    machine = Machine(**metadata)

    with pytest.raises(peewee.DoesNotExist):
        PostgresMachine.get(PostgresMachine.name == machine.name)
    machine.report()
    PostgresMachine.get(PostgresMachine.name == machine.name)



# -*- coding: utf-8 -*-

import pytest

from gordo.machine.model.factories.utils import hourglass_calc_dims, check_dim_func_len


@pytest.mark.parametrize(
    "test_input,test_expected",
    [
        ((0.2, 4, 5), (4, 3, 2, 1)),
        ((0.5, 3, 10), (8, 7, 5)),
        ((0.5, 3, 3), (3, 2, 2)),
        ((0.3, 3, 10), (8, 5, 3)),
        ((1, 3, 10), (10, 10, 10)),
        ((0, 3, 100000), (66667, 33334, 1)),
    ],
)
def test_hourglass_calc_dims_check_dims(test_input, test_expected):
    """
    Test that hourglass_calc_dims implements the correct dimensions
    """
    dims = hourglass_calc_dims(*test_input)
    assert dims == test_expected


def test_check_dim_func_len():
    """
    Test that error is raised if encoding/decoding parameters are not equal of length
    """
    with pytest.raises(ValueError):
        check_dim_func_len("test", dim=(256, 128), func=("tanh", "tanh", "tanh"))

    with pytest.raises(ValueError):
        check_dim_func_len("test", dim=(256, 128, 56), func=("tanh", "tanh"))



# -*- coding: utf-8 -*-

from datetime import timedelta

import pytest
import numpy as np
import pandas as pd
import yaml

from sklearn.preprocessing import MinMaxScaler, RobustScaler
from sklearn.multioutput import MultiOutputRegressor
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import cross_validate
from sklearn.model_selection import TimeSeriesSplit, KFold

from typing import cast

from gordo import serializer
from gordo.machine.model import utils as model_utils
from gordo.machine.model.base import GordoBase
from gordo.machine.model.anomaly.base import AnomalyDetectorBase
from gordo.machine.model.anomaly.diff import (
    DiffBasedAnomalyDetector,
    DiffBasedKFCVAnomalyDetector,
)


@pytest.mark.parametrize("scaler", (MinMaxScaler(), RobustScaler()))
@pytest.mark.parametrize(
    "index", (range(300), pd.date_range("2019-01-01", "2019-01-30", periods=300))
)
@pytest.mark.parametrize("with_thresholds", (True, False))
@pytest.mark.parametrize("shuffle", (True, False))
def test_diff_detector(scaler, index, with_thresholds: bool, shuffle: bool):
    """
    Test the functionality of the DiffBasedAnomalyDetector
    """

    # Some dataset.
    X, y = (
        pd.DataFrame(np.random.random((300, 3))),
        pd.DataFrame(np.random.random((300, 3))),
    )

    base_estimator = MultiOutputRegressor(estimator=LinearRegression())
    model = DiffBasedAnomalyDetector(
        base_estimator=base_estimator,
        scaler=scaler,
        require_thresholds=with_thresholds,
        shuffle=shuffle,
    )

    assert isinstance(model, AnomalyDetectorBase)

    assert model.get_params() == dict(
        base_estimator=base_estimator, scaler=scaler, shuffle=shuffle
    )

    if with_thresholds:
        model.cross_validate(X=X, y=y)

    model.fit(X, y)

    output: np.ndarray = model.predict(X)
    base_df = model_utils.make_base_dataframe(
        tags=["A", "B", "C"], model_input=X, model_output=output, index=index
    )

    # Base prediction dataframe has none of these columns
    assert not any(
        col in base_df.columns
        for col in (
            "total-anomaly-scaled",
            "total-anomaly-unscaled",
            "tag-anomaly-scaled",
            "tag-anomaly-unscaled",
        )
    )

    # Apply the anomaly detection logic on the base prediction df
    anomaly_df = model.anomaly(X, y, timedelta(days=1))

    # Should have these added error calculated columns now.
    assert all(
        col in anomaly_df.columns
        for col in (
            "total-anomaly-scaled",
            "total-anomaly-unscaled",
            "tag-anomaly-scaled",
            "tag-anomaly-unscaled",
        )
    )

    # Verify calculation for unscaled data
    feature_error_unscaled = np.abs(base_df["model-output"].values - y.values)
    total_anomaly_unscaled = np.square(feature_error_unscaled).mean(axis=1)
    assert np.allclose(
        feature_error_unscaled, anomaly_df["tag-anomaly-unscaled"].values
    )
    assert np.allclose(
        total_anomaly_unscaled, anomaly_df["total-anomaly-unscaled"].values
    )

    # Verify calculations for scaled data
    feature_error_scaled = np.abs(
        scaler.transform(base_df["model-output"].values) - scaler.transform(y)
    )
    total_anomaly_scaled = np.square(feature_error_scaled).mean(axis=1)
    assert np.allclose(feature_error_scaled, anomaly_df["tag-anomaly-scaled"].values)
    assert np.allclose(total_anomaly_scaled, anomaly_df["total-anomaly-scaled"].values)

    if with_thresholds:
        assert "anomaly-confidence" in anomaly_df.columns
        assert "total-anomaly-confidence" in anomaly_df.columns
        assert anomaly_df["anomaly-confidence"].notnull().to_numpy().any()
        assert anomaly_df["total-anomaly-confidence"].notnull().to_numpy().all()

    else:
        assert "anomaly-confidence" not in anomaly_df.columns
        assert "total-anomaly-confidence" not in anomaly_df.columns


@pytest.mark.parametrize("scaler", (MinMaxScaler(), RobustScaler()))
@pytest.mark.parametrize("len_x_y", (100, 144, 300))
@pytest.mark.parametrize("time_index", (True, False))
@pytest.mark.parametrize("with_thresholds", (True, False))
@pytest.mark.parametrize("shuffle", (True, False))
@pytest.mark.parametrize("window", (None, 144))
@pytest.mark.parametrize("smoothing_method", (None, "smm", "sma", "ewma"))
def test_diff_detector_with_window(
    scaler,
    len_x_y: int,
    time_index: bool,
    with_thresholds: bool,
    shuffle: bool,
    window,
    smoothing_method,
):
    """
    Test the functionality of the DiffBasedAnomalyDetector with window
    """

    # Some dataset.
    X, y = (
        pd.DataFrame(np.random.random((len_x_y, 3))),
        pd.DataFrame(np.random.random((len_x_y, 3))),
    )
    tags = ["A", "B", "C"]
    if time_index:
        index = pd.date_range("2019-01-01", "2019-01-11", periods=len_x_y)
    else:
        index = range(len_x_y)

    base_estimator = MultiOutputRegressor(estimator=LinearRegression())
    model = DiffBasedAnomalyDetector(
        base_estimator=base_estimator,
        scaler=scaler,
        require_thresholds=with_thresholds,
        shuffle=shuffle,
        window=window,
        smoothing_method=smoothing_method,
    )

    assert isinstance(model, AnomalyDetectorBase)

    if window is None:
        assert model.get_params() == dict(
            base_estimator=base_estimator, scaler=scaler, shuffle=shuffle
        )

    elif window is not None and smoothing_method is None:
        assert model.get_params() == dict(
            base_estimator=base_estimator,
            scaler=scaler,
            shuffle=shuffle,
            window=window,
            smoothing_method="smm",
        )

    else:
        assert model.get_params() == dict(
            base_estimator=base_estimator,
            scaler=scaler,
            shuffle=shuffle,
            window=window,
            smoothing_method=smoothing_method,
        )

    if with_thresholds:
        model.cross_validate(X=X, y=y)

    model.fit(X, y)

    output: np.ndarray = model.predict(X)
    base_df = model_utils.make_base_dataframe(
        tags=tags, model_input=X, model_output=output, index=index
    )
    # Base prediction dataframe has none of these columns
    assert not any(
        col in base_df.columns
        for col in (
            "total-anomaly-scaled",
            "total-anomaly-unscaled",
            "tag-anomaly-scaled",
            "tag-anomaly-unscaled",
            "smooth-total-anomaly-scaled",
            "smooth-total-anomaly-unscaled",
            "smooth-tag-anomaly-scaled",
            "smooth-tag-anomaly-unscaled",
        )
    )

    # Apply the anomaly detection logic on the base prediction df
    anomaly_df = model.anomaly(X, y)
    anomaly_df = cast(pd.DataFrame, anomaly_df)

    # Should have these added error calculated columns now.
    if window is not None:
        assert all(
            col in anomaly_df.columns
            for col in (
                "total-anomaly-scaled",
                "total-anomaly-unscaled",
                "tag-anomaly-scaled",
                "tag-anomaly-unscaled",
                "smooth-total-anomaly-scaled",
                "smooth-total-anomaly-unscaled",
                "smooth-tag-anomaly-scaled",
                "smooth-tag-anomaly-unscaled",
            )
        )
    else:
        assert all(
            col in anomaly_df.columns
            for col in (
                "total-anomaly-scaled",
                "total-anomaly-unscaled",
                "tag-anomaly-scaled",
                "tag-anomaly-unscaled",
            )
        )
        assert not any(
            col in base_df.columns
            for col in (
                "smooth-total-anomaly-scaled",
                "smooth-total-anomaly-unscaled",
                "smooth-tag-anomaly-scaled",
                "smooth-tag-anomaly-unscaled",
            )
        )

    # Verify calculation for unscaled data
    feature_error_unscaled = pd.DataFrame(
        data=np.abs(base_df["model-output"].to_numpy() - y.to_numpy()),
        index=index,
        columns=tags,
    )
    total_anomaly_unscaled = pd.Series(
        data=np.square(feature_error_unscaled).mean(axis=1)
    )
    assert np.allclose(
        feature_error_unscaled.to_numpy(), anomaly_df["tag-anomaly-unscaled"].to_numpy()
    )
    assert np.allclose(
        total_anomaly_unscaled.to_numpy(),
        anomaly_df["total-anomaly-unscaled"].to_numpy(),
    )

    if window is not None:
        if smoothing_method is None or smoothing_method == "smm":
            smooth_feature_error_unscaled = (
                feature_error_unscaled.rolling(model.window).median().dropna()
            )
            smooth_total_anomaly_unscaled = (
                total_anomaly_unscaled.rolling(model.window).median().dropna()
            )

        elif smoothing_method == "sma":
            smooth_feature_error_unscaled = (
                feature_error_unscaled.rolling(model.window).mean().dropna()
            )
            smooth_total_anomaly_unscaled = (
                total_anomaly_unscaled.rolling(model.window).mean().dropna()
            )
        elif smoothing_method == "ewma":
            smooth_feature_error_unscaled = feature_error_unscaled.ewm(
                span=model.window
            ).mean()
            smooth_total_anomaly_unscaled = total_anomaly_unscaled.ewm(
                span=model.window
            ).mean()

        assert np.allclose(
            smooth_feature_error_unscaled.to_numpy(),
            anomaly_df["smooth-tag-anomaly-unscaled"].dropna().to_numpy(),
        )
        assert np.allclose(
            smooth_total_anomaly_unscaled.to_numpy(),
            anomaly_df["smooth-total-anomaly-unscaled"].dropna().to_numpy(),
        )

    # Verify calculations for scaled data
    feature_error_scaled = pd.DataFrame(
        data=np.abs(
            scaler.transform(base_df["model-output"].to_numpy()) - scaler.transform(y)
        ),
        index=index,
        columns=tags,
    )
    total_anomaly_scaled = pd.Series(data=np.square(feature_error_scaled).mean(axis=1))
    assert np.allclose(
        feature_error_scaled.to_numpy(), anomaly_df["tag-anomaly-scaled"].to_numpy()
    )
    assert np.allclose(
        total_anomaly_scaled, anomaly_df["total-anomaly-scaled"].to_numpy()
    )

    if window is not None:
        if smoothing_method is None or smoothing_method == "smm":
            smooth_feature_error_scaled = (
                feature_error_scaled.rolling(model.window).median().dropna()
            )
            smooth_total_anomaly_scaled = (
                total_anomaly_scaled.rolling(model.window).median().dropna()
            )
        elif smoothing_method == "sma":
            smooth_feature_error_scaled = (
                feature_error_scaled.rolling(model.window).mean().dropna()
            )
            smooth_total_anomaly_scaled = (
                total_anomaly_scaled.rolling(model.window).mean().dropna()
            )
        elif smoothing_method == "ewma":
            smooth_feature_error_scaled = feature_error_scaled.ewm(
                span=model.window
            ).mean()
            smooth_total_anomaly_scaled = total_anomaly_scaled.ewm(
                span=model.window
            ).mean()

        assert np.allclose(
            smooth_feature_error_scaled.to_numpy(),
            anomaly_df["smooth-tag-anomaly-scaled"].dropna().to_numpy(),
        )
        assert np.allclose(
            smooth_total_anomaly_scaled.to_numpy(),
            anomaly_df["smooth-total-anomaly-scaled"].dropna().to_numpy(),
        )

    # Check number of NA's is consistent with window size
    if (
        smoothing_method != "ewma"
        and model.window is not None
        and len_x_y >= model.window
    ):
        assert (
            anomaly_df["smooth-tag-anomaly-scaled"].isna().sum().sum()
            == (model.window - 1) * anomaly_df["smooth-tag-anomaly-scaled"].shape[1]
        )
        assert (
            anomaly_df["smooth-total-anomaly-scaled"].isna().sum() == model.window - 1
        )

    if with_thresholds:
        assert "anomaly-confidence" in anomaly_df.columns
        assert "total-anomaly-confidence" in anomaly_df.columns
        assert anomaly_df["anomaly-confidence"].notnull().to_numpy().all()
        assert anomaly_df["total-anomaly-confidence"].notnull().to_numpy().all()
    else:
        assert "anomaly-confidence" not in anomaly_df.columns
        assert "total-anomaly-confidence" not in anomaly_df.columns


@pytest.mark.parametrize("scaler", (MinMaxScaler(), RobustScaler()))
@pytest.mark.parametrize(
    "index", (range(300), pd.date_range("2019-01-01", "2019-01-30", periods=300))
)
@pytest.mark.parametrize("with_thresholds", (True, False))
@pytest.mark.parametrize("shuffle", (True, False))
@pytest.mark.parametrize("window", (6, 144))
@pytest.mark.parametrize("smoothing_method", ("smm", "sma", "ewma"))
@pytest.mark.parametrize("threshold_percentile", (0.975, 1.0))
def test_diff_kfcv_detector(
    scaler,
    index,
    with_thresholds: bool,
    shuffle: bool,
    window: int,
    smoothing_method: str,
    threshold_percentile: float,
):
    """
    Test the functionality of the DiffBasedKFCVAnomalyDetector
    """

    # Some dataset.
    X, y = (
        pd.DataFrame(np.random.random((300, 3))),
        pd.DataFrame(np.random.random((300, 3))),
    )

    base_estimator = MultiOutputRegressor(estimator=LinearRegression())
    model = DiffBasedKFCVAnomalyDetector(
        base_estimator=base_estimator,
        scaler=scaler,
        require_thresholds=with_thresholds,
        shuffle=shuffle,
        window=window,
        smoothing_method=smoothing_method,
        threshold_percentile=threshold_percentile,
    )

    assert isinstance(model, AnomalyDetectorBase)

    assert model.get_params() == dict(
        base_estimator=base_estimator,
        scaler=scaler,
        window=window,
        smoothing_method=smoothing_method,
        shuffle=shuffle,
        threshold_percentile=threshold_percentile,
    )

    if with_thresholds:
        model.cross_validate(X=X, y=y)

    model.fit(X, y)

    output: np.ndarray = model.predict(X)
    base_df = model_utils.make_base_dataframe(
        tags=["A", "B", "C"], model_input=X, model_output=output, index=index
    )

    # Base prediction dataframe has none of these columns
    assert not any(
        col in base_df.columns
        for col in (
            "total-anomaly-scaled",
            "total-anomaly-unscaled",
            "tag-anomaly-scaled",
            "tag-anomaly-unscaled",
        )
    )

    # Apply the anomaly detection logic on the base prediction df
    anomaly_df = model.anomaly(X, y, timedelta(days=1))

    # Should have these added error calculated columns now.
    assert all(
        col in anomaly_df.columns
        for col in (
            "total-anomaly-scaled",
            "total-anomaly-unscaled",
            "tag-anomaly-scaled",
            "tag-anomaly-unscaled",
        )
    )

    # Verify calculation for unscaled data
    feature_error_unscaled = np.abs(base_df["model-output"].values - y.values)
    total_anomaly_unscaled = np.square(feature_error_unscaled).mean(axis=1)
    assert np.allclose(
        feature_error_unscaled, anomaly_df["tag-anomaly-unscaled"].values
    )
    assert np.allclose(
        total_anomaly_unscaled, anomaly_df["total-anomaly-unscaled"].values
    )

    # Verify calculations for scaled data
    feature_error_scaled = np.abs(
        scaler.transform(base_df["model-output"].values) - scaler.transform(y)
    )
    total_anomaly_scaled = np.square(feature_error_scaled).mean(axis=1)
    assert np.allclose(feature_error_scaled, anomaly_df["tag-anomaly-scaled"].values)
    assert np.allclose(total_anomaly_scaled, anomaly_df["total-anomaly-scaled"].values)

    if with_thresholds:
        metadata = model.get_metadata()
        assert not any(np.isnan(metadata["feature-thresholds"]))
        assert not np.isnan(metadata["aggregate-threshold"])
        assert "anomaly-confidence" in anomaly_df.columns
        assert "total-anomaly-confidence" in anomaly_df.columns
        assert anomaly_df["anomaly-confidence"].notnull().to_numpy().all()
        assert anomaly_df["total-anomaly-confidence"].notnull().to_numpy().all()
    else:
        assert "anomaly-confidence" not in anomaly_df.columns
        assert "total-anomaly-confidence" not in anomaly_df.columns


@pytest.mark.parametrize(
    "config",
    (
        """
        gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
            base_estimator:
                sklearn.pipeline.Pipeline:
                    steps:
                        - sklearn.preprocessing.MinMaxScaler
                        - gordo.machine.model.models.KerasAutoEncoder:
                            kind: feedforward_hourglass
                    memory:
    """,
        """
        gordo.machine.model.anomaly.diff.DiffBasedAnomalyDetector:
            base_estimator:
                gordo.machine.model.models.KerasAutoEncoder:
                    kind: feedforward_hourglass
    """,
        """
        gordo.machine.model.anomaly.diff.DiffBasedKFCVAnomalyDetector:
            base_estimator:
                sklearn.compose.TransformedTargetRegressor:
                    transformer:
                        sklearn.preprocessing.MinMaxScaler
                    regressor:
                        sklearn.pipeline.Pipeline:
                            steps:
                            - sklearn.preprocessing.MinMaxScaler
                            - gordo.machine.model.models.KerasAutoEncoder:
                                kind: feedforward_hourglass
                                batch_size: 128
                                compression_factor: 0.5
                                encoding_layers: 1
                                func: tanh
                                out_func: linear
                                optimizer: Adam
                                loss: mse
                                epochs: 1000
                                validation_split: 0.1
                                callbacks:
                                    - tensorflow.keras.callbacks.EarlyStopping:
                                        monitor: val_loss
                                        patience: 10
                                        restore_best_weights: true
            scaler: sklearn.preprocessing.MinMaxScaler
            window: 144
            shuffle: true
            threshold_percentile: 0.975
    """,
    ),
)
def test_diff_detector_serializability(config):
    """
    Should play well with the gordo serializer
    """
    config = yaml.safe_load(config)

    model = serializer.from_definition(config)
    serializer.into_definition(model)
    serialized_bytes = serializer.dumps(model)
    serializer.loads(serialized_bytes)


@pytest.mark.parametrize("n_features_y", range(1, 3))
@pytest.mark.parametrize("n_features_x", range(1, 3))
@pytest.mark.parametrize("mode", ("tscv", "kfcv"))
def test_diff_detector_threshold(mode: str, n_features_x: int, n_features_y: int):
    """
    Basic construction logic of thresholds_ attribute in the
    DiffBasedAnomalyDetector and DiffBasedKFCVAnomalyDetector
    """
    X = np.random.random((300, n_features_x))
    y = np.random.random((300, n_features_y))

    base_estimator = MultiOutputRegressor(estimator=LinearRegression())
    if mode == "tscv":
        model = DiffBasedAnomalyDetector(base_estimator=base_estimator)
    elif mode == "kfcv":
        model = DiffBasedKFCVAnomalyDetector(base_estimator=base_estimator)

    # Model has own implementation of cross_validate
    assert hasattr(model, "cross_validate")

    # When initialized it should not have a threshold calculated.
    assert not hasattr(model, "feature_thresholds_")
    assert not hasattr(model, "aggregate_threshold_")
    assert not hasattr(model, "feature_thresholds_per_fold_")
    assert not hasattr(model, "aggregate_thresholds_per_fold_")

    model.fit(X, y)

    # Until it has done cross validation, it has no threshold.
    assert not hasattr(model, "feature_thresholds_")
    assert not hasattr(model, "aggregate_threshold_")
    assert not hasattr(model, "feature_thresholds_per_fold_")
    assert not hasattr(model, "aggregate_thresholds_per_fold_")

    # Calling cross validate should set the threshold for it.
    model.cross_validate(X=X, y=y)

    # Now we have calculated thresholds based on cross validation folds
    assert hasattr(model, "feature_thresholds_")
    assert hasattr(model, "aggregate_threshold_")
    assert isinstance(model.feature_thresholds_, pd.Series)
    assert len(model.feature_thresholds_) == y.shape[1]
    assert all(model.feature_thresholds_.notna())

    if not isinstance(model, DiffBasedKFCVAnomalyDetector):
        assert hasattr(model, "feature_thresholds_per_fold_")
        assert hasattr(model, "aggregate_thresholds_per_fold_")
        assert isinstance(model.feature_thresholds_per_fold_, pd.DataFrame)
        assert isinstance(model.aggregate_thresholds_per_fold_, dict)


@pytest.mark.parametrize("n_features_y", range(1, 3))
@pytest.mark.parametrize("n_features_x", range(1, 3))
@pytest.mark.parametrize("len_x_y", [100, 144, 1440])
def test_diff_detector_threshold_with_window(
    n_features_y: int, n_features_x: int, len_x_y: int
):
    """
    Basic construction logic of thresholds_ attribute in the
    DiffBasedAnomalyDetector
    """
    X = np.random.random((len_x_y, n_features_x))
    y = np.random.random((len_x_y, n_features_y))

    model = DiffBasedAnomalyDetector(
        base_estimator=MultiOutputRegressor(estimator=LinearRegression()), window=144
    )

    # Model has own implementation of cross_validate
    assert hasattr(model, "cross_validate")

    # When initialized it should not have a threshold calculated.
    assert not hasattr(model, "feature_thresholds_")
    assert not hasattr(model, "aggregate_threshold_")
    assert not hasattr(model, "feature_thresholds_per_fold_")
    assert not hasattr(model, "aggregate_thresholds_per_fold_")
    assert not hasattr(model, "smooth_feature_thresholds_")
    assert not hasattr(model, "smooth_aggregate_threshold_")
    assert not hasattr(model, "smooth_feature_thresholds_per_fold_")
    assert not hasattr(model, "smooth_aggregate_thresholds_per_fold_")

    model.fit(X, y)

    # Until it has done cross validation, it has no threshold.
    assert not hasattr(model, "feature_thresholds_")
    assert not hasattr(model, "aggregate_threshold_")
    assert not hasattr(model, "feature_thresholds_per_fold_")
    assert not hasattr(model, "aggregate_thresholds_per_fold_")
    assert not hasattr(model, "smooth_feature_thresholds_")
    assert not hasattr(model, "smooth_aggregate_threshold_")
    assert not hasattr(model, "smooth_feature_thresholds_per_fold_")
    assert not hasattr(model, "smooth_aggregate_thresholds_per_fold_")

    # Calling cross validate should set the threshold for it.
    model.cross_validate(X=X, y=y)

    # Now we have calculated thresholds based on cross validation folds
    assert hasattr(model, "feature_thresholds_")
    assert hasattr(model, "aggregate_threshold_")
    assert hasattr(model, "feature_thresholds_per_fold_")
    assert hasattr(model, "aggregate_thresholds_per_fold_")
    assert isinstance(model.feature_thresholds_, pd.Series)
    assert len(model.feature_thresholds_) == y.shape[1]
    assert all(model.feature_thresholds_.notna())
    assert isinstance(model.feature_thresholds_per_fold_, pd.DataFrame)
    assert isinstance(model.aggregate_thresholds_per_fold_, dict)

    assert hasattr(model, "smooth_feature_thresholds_")
    assert hasattr(model, "smooth_aggregate_threshold_")
    assert hasattr(model, "smooth_feature_thresholds_per_fold_")
    assert hasattr(model, "smooth_aggregate_thresholds_per_fold_")
    assert isinstance(model.smooth_feature_thresholds_, pd.Series)
    assert len(model.smooth_feature_thresholds_) == y.shape[1]
    if model.window is not None and len_x_y <= model.window:
        assert all(model.smooth_feature_thresholds_.isna())
    else:
        assert all(model.smooth_feature_thresholds_.notna())
    assert isinstance(model.smooth_feature_thresholds_per_fold_, pd.DataFrame)
    assert isinstance(model.smooth_aggregate_thresholds_per_fold_, dict)


@pytest.mark.parametrize("mode", ("tscv", "tscv_win", "kfcv"))
def test_diff_detector_get_metadata(mode):
    """
    Test if metadata is generated as expected.
    """
    X = pd.DataFrame(np.random.random((200, 5)))
    y = pd.DataFrame(np.random.random((200, 2)))

    base_estimator = MultiOutputRegressor(LinearRegression())
    if mode == "tscv":
        model = DiffBasedAnomalyDetector(base_estimator=base_estimator)
    elif mode == "tscv_win":
        model = DiffBasedAnomalyDetector(base_estimator=base_estimator, window=144)
    elif mode == "kfcv":
        model = DiffBasedKFCVAnomalyDetector(base_estimator=base_estimator)

    metadata = model.get_metadata()
    assert isinstance(metadata, dict)

    if not isinstance(model, GordoBase):
        assert "base_estimator" in metadata.keys()
        assert "scaler" in metadata.keys()
        assert "shuffle" in metadata.keys()

    # When initialized it should not have a threshold calculated.
    assert "feature-thresholds" not in metadata.keys()
    assert "aggregate-threshold" not in metadata.keys()
    assert "feature-thresholds-per-fold" not in metadata.keys()
    assert "aggregate-thresholds-per-fold" not in metadata.keys()

    # Calling cross validate should set the threshold for it.
    model.cross_validate(X=X, y=y)
    metadata = model.get_metadata()

    if not isinstance(model, GordoBase):
        assert "base_estimator" in metadata.keys()
        assert "scaler" in metadata.keys()
        assert "shuffle" in metadata.keys()

    # Now we have calculated thresholds based on cross validation folds
    assert "feature-thresholds" in metadata.keys()
    assert "aggregate-threshold" in metadata.keys()
    assert isinstance(metadata["feature-thresholds"], list)
    assert len(metadata["feature-thresholds"]) == y.shape[1]

    if mode != "kfcv":
        assert "feature-thresholds-per-fold" in metadata.keys()
        assert "aggregate-thresholds-per-fold" in metadata.keys()
    if mode != "tscv":
        assert "window" in metadata.keys()
        assert "smoothing-method" in metadata.keys()
    if mode == "tscv_win":
        assert "smooth-feature-thresholds" in metadata.keys()
        assert "smooth-aggregate-threshold" in metadata.keys()
        assert "smooth-feature-thresholds-per-fold" in metadata.keys()
        assert "smooth-aggregate-thresholds-per-fold" in metadata.keys()
    if mode == "kfcv":
        assert "threshold-percentile" in metadata.keys()


@pytest.mark.parametrize("return_estimator", (True, False))
@pytest.mark.parametrize("n_features_y", (1, 10))
@pytest.mark.parametrize("mode", ("tscv", "kfcv"))
def test_diff_detector_cross_validate(
    mode: str, n_features_y: int, return_estimator: bool
):
    """
    DiffBasedAnomalyDetector.cross_validate implementation should be the
    same as sklearn.model_selection.cross_validate if called the same.

    And it always will update `return_estimator` to True, as it requires
    the intermediate models to calculate the thresholds
    """
    X = np.random.random((100, 10))
    y = np.random.random((100, n_features_y))

    base_estimator = MultiOutputRegressor(LinearRegression())
    if mode == "tscv":
        model = DiffBasedAnomalyDetector(base_estimator=base_estimator)
        cv = TimeSeriesSplit(n_splits=3)
    elif mode == "kfcv":
        model = DiffBasedKFCVAnomalyDetector(base_estimator=base_estimator)
        cv = KFold(n_splits=5, shuffle=True, random_state=0)

    cv_results_da = model.cross_validate(
        X=X, y=y, cv=cv, return_estimator=return_estimator
    )
    cv_results_sk = cross_validate(model, X=X, y=y, cv=cv, return_estimator=True)

    assert cv_results_da.keys() == cv_results_sk.keys()


@pytest.mark.parametrize("require_threshold", (True, False))
@pytest.mark.parametrize("mode", ("tscv", "kfcv"))
def test_diff_detector_require_thresholds(mode: str, require_threshold: bool):
    """
    Should fail if requiring thresholds, but not calling cross_validate
    """
    X = pd.DataFrame(np.random.random((200, 5)))
    y = pd.DataFrame(np.random.random((200, 2)))

    base_estimator = MultiOutputRegressor(LinearRegression())
    if mode == "tscv":
        model = DiffBasedAnomalyDetector(
            base_estimator=base_estimator, require_thresholds=require_threshold
        )
    elif mode == "kfcv":
        model = DiffBasedKFCVAnomalyDetector(
            base_estimator=base_estimator, require_thresholds=require_threshold
        )
    model.fit(X, y)

    if require_threshold:
        # FAIL: Forgot to call .cross_validate to calculate thresholds.
        with pytest.raises(AttributeError):
            model.anomaly(X, y)

        model.cross_validate(X=X, y=y)
        model.anomaly(X, y)
    else:
        # thresholds not required
        model.anomaly(X, y)



import re
from typing import Optional, Union

from dataclasses import dataclass
from enum import Enum
from abc import ABCMeta, abstractmethod


class Version(metaclass=ABCMeta):
    @abstractmethod
    def get_version(self):
        ...


class Special(Enum):
    LATEST = "latest"
    STABLE = "stable"

    @classmethod
    def find(cls, version: str) -> Optional["Special"]:
        for special in cls:
            if special.value == version:
                return special
        return None


@dataclass(frozen=True)
class GordoSpecial(Version):
    special: Special

    def get_version(self):
        return self.special.value


release_re = re.compile(r"^(\d{1,5})(\.(\d+)(\.(\d+)(.*?)?)?)?$")


@dataclass(frozen=True)
class GordoRelease(Version):
    major: int
    minor: Optional[int] = None
    patch: Optional[int] = None
    suffix: Optional[str] = None

    def without_patch(self) -> bool:
        return self.suffix is None and self.patch is None

    def only_major_minor(self) -> bool:
        return (
            self.major is not None and self.minor is not None and self.without_patch()
        )

    def only_major(self) -> bool:
        return self.major is not None and self.minor is None and self.without_patch()

    def get_version(self):
        version = str(self.major)
        if self.minor is not None:
            version += "." + str(self.minor)
        if self.patch is not None:
            version += "." + str(self.patch)
        if self.suffix is not None:
            version += self.suffix
        return version


pr_prefix = "pr-"


@dataclass(frozen=True)
class GordoPR(Version):
    number: int

    def get_version(self):
        return "%s%d" % (pr_prefix, self.number)


sha_re = re.compile(r"^[0-9a-z]{8,40}$")


@dataclass(frozen=True)
class GordoSHA(Version):
    sha: str

    def get_version(self):
        return self.sha


def parse_version(
    gordo_version: str,
) -> Union[GordoRelease, GordoSpecial, GordoPR, GordoSHA]:
    """
    Parsing gordo version. Also supported gordo docker images tags

    Parameters
    ----------
    gordo_version: str

    Example
    -------
    >>> parse_version('2.3.5')
    GordoRelease(major=2, minor=3, patch=5, suffix=None)
    >>> parse_version('latest')
    GordoSpecial(special=<Special.LATEST: 'latest'>)

    Returns
    -------
    Union[GordoRelease, GordoSpecial, GordoPR, GordoSHA]

    """
    special_version = Special.find(gordo_version)
    if special_version is not None:
        return GordoSpecial(special_version)
    if gordo_version.find(pr_prefix) == 0:
        try:
            number = int(gordo_version[len(pr_prefix) :])
        except ValueError:
            raise ValueError("Malformed gordo PR version '%s'" % gordo_version)
        return GordoPR(number)
    m = release_re.match(gordo_version)
    if m:
        (major, _, minor, _, patch, suffix) = m.groups()
        return GordoRelease(
            int(major),
            int(minor) if minor else None,
            int(patch) if patch else None,
            suffix if suffix else None,
        )
    m = sha_re.match(gordo_version)
    if m:
        return GordoSHA(gordo_version)
    raise ValueError("Malformed gordo version '%s'" % gordo_version)



# -*- coding: utf-8 -*-

from flask import url_for
from flask_restplus import Api as BaseApi


class Api(BaseApi):
    """
    Redefine Api to ensure specs url is relative
    """

    @property
    def specs_url(self):
        return url_for(self.endpoint("specs"), _external=False)



# -*- coding: utf-8 -*-

import unittest
from gordo.machine.model.register import register_model_builder


class RegisterTestCase(unittest.TestCase):
    def test_fail_no_required_params(self):
        """
        @register required a function which has certain parameters.

        n_features
        """
        # Fail without required param(s)
        with self.assertRaises(ValueError):

            @register_model_builder(type="KerasAutoEncoder")
            def build_fn():
                pass

        # Pass with required param(s)
        @register_model_builder(type="KerasAutoEncoder")  # pragma: no flakes
        def build_fn(n_features):
            pass

        # Call to ensure that register didn't 'eat' the function
        build_fn(1)

    def test_hold_multiple_funcs(self):
        """
        Ensure the register holds references to multiple funcs
        """

        @register_model_builder(type="KerasAutoEncoder")
        def func1(n_features):
            pass

        @register_model_builder(type="KerasAutoEncoder")
        def func2(n_features):
            pass

        self.assertTrue(
            all(
                func_name in register_model_builder.factories["KerasAutoEncoder"]
                for func_name in ["func1", "func2"]
            )
        )



from typing import Dict, Any, Optional
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from gordo import __version__


__all__ = [
    "Metadata",
    "BuildMetadata",
    "ModelBuildMetadata",
    "CrossValidationMetaData",
    "DatasetBuildMetadata",
]


@dataclass_json
@dataclass
class CrossValidationMetaData:
    scores: Dict[str, Any] = field(default_factory=dict)
    cv_duration_sec: Optional[float] = None
    splits: Dict[str, Any] = field(default_factory=dict)


@dataclass_json
@dataclass
class ModelBuildMetadata:
    model_offset: int = 0
    model_creation_date: Optional[str] = None
    model_builder_version: str = __version__
    cross_validation: CrossValidationMetaData = field(
        default_factory=CrossValidationMetaData
    )
    model_training_duration_sec: Optional[float] = None
    model_meta: Dict[str, Any] = field(default_factory=dict)


@dataclass_json
@dataclass
class DatasetBuildMetadata:
    query_duration_sec: Optional[float] = None  # How long it took to get the data
    dataset_meta: Dict[str, Any] = field(default_factory=dict)


@dataclass_json
@dataclass
class BuildMetadata:
    model: ModelBuildMetadata = field(default_factory=ModelBuildMetadata)
    dataset: DatasetBuildMetadata = field(default_factory=DatasetBuildMetadata)


@dataclass_json
@dataclass
class Metadata:
    user_defined: Dict[str, Any] = field(default_factory=dict)
    build_metadata: BuildMetadata = field(default_factory=BuildMetadata)



# -*- coding: utf-8 -*-

import dateutil.parser
import glob
import importlib
import logging
import mock
import nbformat
import os
import sys

import pandas as pd
import numpy as np

from nbconvert.exporters import PythonExporter

from gordo_dataset.data_provider.providers import DataLakeProvider
from gordo_dataset.datasets import TimeSeriesDataset
from gordo_dataset.sensor_tag import SensorTag

logger = logging.getLogger(__name__)

CONFIG = dict(
    train_start_date=dateutil.parser.isoparse("2014-07-01T00:10:00+00:00"),
    train_end_date=dateutil.parser.isoparse("2015-01-01T00:00:00+00:00"),
    tag_list=[SensorTag(name=f"Tag {i}", asset=None) for i in range(10)],
)


def _fake_data():
    data = pd.DataFrame(
        {
            f"sensor_{i}": np.random.random(size=100)
            for i in range(len(CONFIG["tag_list"]))
        }
    )
    return data, data


@mock.patch.object(TimeSeriesDataset, "get_data", return_value=_fake_data())
def test_faked_DataLakeBackedDataset(MockDataset):

    provider = DataLakeProvider(storename="dataplatformdlsprod", interactive=True)
    dataset = TimeSeriesDataset(data_provider=provider, **CONFIG)

    # Should be able to call get_data without being asked to authenticate in tests
    X, y = dataset.get_data()


@mock.patch.object(TimeSeriesDataset, "get_data", return_value=_fake_data())
def test_notebooks(MockDataset, tmpdir):
    """
    Ensures all notebooks will run without error
    """
    repo_root = os.path.join(os.path.dirname(__file__), "..")
    notebooks = glob.glob(os.path.join(repo_root, "examples", "*.ipynb"))

    logger.info(f"Found {len(notebooks)} notebooks to test.")

    for notebook in notebooks:

        logger.info(f"Testing notebook: {os.path.basename(notebook)}")

        with open(notebook) as f:
            nb = nbformat.read(f, as_version=4)
            exporter = PythonExporter()
            source, _meta = exporter.from_notebook_node(nb)

            with open(os.path.join(tmpdir, "tmpmodule.py"), "w") as f:
                f.writelines(source)
            with open(os.path.join(tmpdir, "__init__.py"), "w") as f:
                f.write("from .tmpmodule import *")

            # Import this module to 'run' the code
            module_dir = os.path.join(tmpdir, "..")
            sys.path.insert(0, module_dir)

            importlib.import_module(os.path.basename(tmpdir), ".")

            sys.path.remove(module_dir)



# -*- coding: utf-8 -*-
import json
import logging
from datetime import datetime
from typing import Dict, Any, Union, Optional

import numpy as np
import yaml

from gordo_dataset.base import GordoBaseDataset
from gordo.machine.validators import (
    ValidUrlString,
    ValidMetadata,
    ValidModel,
    ValidDataset,
    ValidMachineRuntime,
)
from gordo.machine.metadata import Metadata
from gordo.workflow.workflow_generator.helpers import patch_dict


logger = logging.getLogger(__name__)


class Machine:
    """
    Represents a single machine in a config file
    """

    name = ValidUrlString()
    project_name = ValidUrlString()
    host = ValidUrlString()
    model = ValidModel()
    dataset = ValidDataset()
    metadata = ValidMetadata()
    runtime = ValidMachineRuntime()
    _strict = True

    def __init__(
        self,
        name: str,
        model: dict,
        dataset: Union[GordoBaseDataset, dict],
        project_name: str,
        evaluation: Optional[dict] = None,
        metadata: Optional[Union[dict, Metadata]] = None,
        runtime=None,
    ):

        if runtime is None:
            runtime = dict()
        if evaluation is None:
            evaluation = dict(cv_mode="full_build")
        if metadata is None:
            metadata = dict()
        self.name = name
        self.model = model
        self.dataset = (
            dataset
            if isinstance(dataset, GordoBaseDataset)
            else GordoBaseDataset.from_dict(dataset)
        )
        self.runtime = runtime
        self.evaluation = evaluation
        self.metadata = (
            metadata
            if isinstance(metadata, Metadata)
            else Metadata.from_dict(metadata)  # type: ignore
        )
        self.project_name = project_name

        self.host = f"gordoserver-{self.project_name}-{self.name}"

    @classmethod
    def from_config(  # type: ignore
        cls, config: Dict[str, Any], project_name: str, config_globals=None
    ):
        """
        Construct an instance from a block of YAML config file which represents
        a single Machine; loaded as a ``dict``.

        Parameters
        ----------
        config: dict
            The loaded block of config which represents a 'Machine' in YAML
        project_name: str
            Name of the project this Machine belongs to.
        config_globals:
            The block of config within the YAML file within `globals`

        Returns
        -------
        :class:`~Machine`
        """
        if config_globals is None:
            config_globals = dict()

        name = config["name"]
        model = config.get("model") or config_globals.get("model")

        local_runtime = config.get("runtime", dict())
        runtime = patch_dict(config_globals.get("runtime", dict()), local_runtime)

        dataset_config = patch_dict(
            config.get("dataset", dict()), config_globals.get("dataset", dict())
        )
        dataset = GordoBaseDataset.from_dict(dataset_config)
        evaluation = patch_dict(
            config_globals.get("evaluation", dict()), config.get("evaluation", dict())
        )

        metadata = Metadata(
            user_defined={
                "global-metadata": config_globals.get("metadata", dict()),
                "machine-metadata": config.get("metadata", dict()),
            }
        )
        return cls(
            name,
            model,
            dataset,
            metadata=metadata,
            runtime=runtime,
            project_name=project_name,
            evaluation=evaluation,
        )

    def __str__(self):
        return yaml.dump(self.to_dict())

    def __eq__(self, other):
        return self.to_dict() == other.to_dict()

    @classmethod
    def from_dict(cls, d: dict) -> "Machine":
        """
        Get an instance from a dict taken from :func:`~Machine.to_dict`
        """
        # No special treatment required, just here for consistency.
        return cls(**d)

    def to_dict(self):
        """
        Convert to a ``dict`` representation along with all attributes which
        can also be converted to a ``dict``. Can reload with :func:`~Machine.from_dict`
        """
        return {
            "name": self.name,
            "dataset": self.dataset.to_dict(),
            "model": self.model,
            "metadata": self.metadata.to_dict(),
            "runtime": self.runtime,
            "project_name": self.project_name,
            "evaluation": self.evaluation,
        }

    def report(self):
        """ 
        Run any reporters in the machine's runtime for the current state.

        Reporters implement the :class:`gordo.reporters.base.BaseReporter` and
        can be specified in a config file of the machine for example:

        .. code-block:: yaml

            runtime:
              reporters:
                - gordo.reporters.postgres.PostgresReporter:
                    host: my-special-host

        """
        # Avoid circular dependency with reporters which import Machine
        from gordo.reporters.base import BaseReporter

        for reporter in map(BaseReporter.from_dict, self.runtime.get("reporters", [])):
            logger.debug(f"Using reporter: {reporter}")
            reporter.report(self)


class MachineEncoder(json.JSONEncoder):
    """
    A JSONEncoder for machine objects, handling datetime.datetime objects as strings
    and handles any numpy numeric instances; both of which common in the ``dict``
    representation of a :class:`~gordo.machine.Machine`

    Example
    -------
    >>> from pytz import UTC
    >>> s = json.dumps({"now":datetime.now(tz=UTC)}, cls=MachineEncoder, indent=4)
    >>> s = '{"now": "2019-11-22 08:34:41.636356+"}'
    """

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S.%f+%z")
        # Typecast builtin and numpy ints and floats to builtin types
        elif np.issubdtype(type(obj), np.floating):
            return float(obj)
        elif np.issubdtype(type(obj), np.integer):
            return int(obj)
        else:
            return json.JSONEncoder.default(self, obj)



class DefinitionTestModel:
    @classmethod
    def from_definition(cls, definition: dict):
        return cls(int(definition.get("depth", 10)))

    def __init__(self, depth):
        self.depth = depth

    def into_definition(self):
        return {"depth": self.depth}



import abc

from typing import Dict, Any

from gordo.machine import Machine
from gordo import serializer


class BaseReporter(abc.ABC):
    @abc.abstractmethod
    def report(self, machine: Machine):
        """Report/log the machine"""

    def get_params(self, deep=False):
        return self._params.copy()

    def to_dict(self) -> dict:
        """
        Serialize this object into a dict representation, which can be used to
        initialize a new object after popping 'type' from the dict.

        Returns
        -------
        dict
        """
        return serializer.into_definition(self)

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> "BaseReporter":
        """
        Reconstruct the reporter from a dict representation or a single
        import path if it doesn't require any init parameters.
        """
        return serializer.from_definition(config)



import os
import requests
import subprocess
import yaml
from collections import namedtuple

import numpy as np
import jinja2
import click
from locust import HttpLocust

from gordo.client.utils import EndpointMetadata

TEMPLATE_FILE = "task_set.py.jinja2"
Task = namedtuple("Task", "name path json")


def fetch_metadata(project_name, host, port, ambassador, watchman_port):
    if ambassador:
        data = requests.get(
            f"{host}:{port}/gordo/v0/{project_name}/", proxies={"http": None}
        )
    else:
        data = requests.get(f"{host}:{watchman_port}", proxies={"http": None})

    json_data = yaml.safe_load(data.content)

    endpoint_metadata = [
        EndpointMetadata(endpoint) for endpoint in json_data["endpoints"]
    ]
    tags = {
        metadata.name: {
            "tag_list": len(metadata.tag_list),
            "target_tag_list": len(metadata.target_tag_list),
        }
        for metadata in endpoint_metadata
    }

    return tags


def generate_random_data(len_x, len_y, samples=100):
    if len_y > 0:
        return {
            "X": np.random.random((samples, len_x)).tolist(),
            "y": np.random.random((samples, len_y)).tolist(),
        }
    return {"X": np.random.random((samples, len_x)).tolist()}


def make_tasks(tags, endpoint):
    tasks = list()
    for name, dict_values in tags.items():
        data = generate_random_data(
            dict_values["tag_list"], dict_values["target_tag_list"]
        )
        path = f"/gordo/v0/{endpoint}/{name}/prediction"
        tasks.append(Task(name=f"fn_{name.replace('-', '_')}", path=path, json=data))

    return tasks


@click.command()
@click.option("--project-name", type=str, help="Project name", required=True)
@click.option(
    "--host",
    type=str,
    default="http://127.0.0.1",
    help="Host is usually 127.0.0.1 (localhost)",
)
@click.option("--port", type=int, default=8888, help="Port which hosts the project")
@click.option(
    "--ambassador", is_flag=True, help="Use if routing should go through ambassador"
)
@click.option(
    "--watchman-port",
    type=int,
    default=8889,
    help="Use if you wanna port-forward watchman to a different port",
)
def main(project_name, host, port, ambassador, watchman_port):
    template_loader = jinja2.FileSystemLoader(searchpath=os.path.dirname(__file__))
    template_env = jinja2.Environment(loader=template_loader)
    template = template_env.get_template(TEMPLATE_FILE)

    tags = fetch_metadata(project_name, host, port, ambassador, watchman_port)
    tasks = make_tasks(tags, project_name)
    path = os.path.join(os.path.dirname(__file__), "task_set.py")
    template.stream(tasks=tasks).dump(path)

    subprocess.run(["locust", "-f", f"{__file__}", "--host", f"{host}:{port}"])


if __name__ == "__main__":
    main()


class MyLocust(HttpLocust):
    from benchmarks.load_test.task_set import Tasks

    task_set = Tasks
    min_wait = 1000
    max_wait = 1000



# -*- coding: utf-8 -*-

import pytest
import numpy as np
from gordo.server import utils as server_utils


@pytest.mark.parametrize(
    # Nested records and single record
    "data_size",
    [10, 1],
)
@pytest.mark.parametrize("resp_format", ("json", "parquet", None))
def test_anomaly_prediction_endpoint(
    base_route,
    sensors_str,
    influxdb,
    gordo_ml_server_client,
    data_size,
    sensors,
    resp_format,
):
    """
    Anomaly GET and POST responses are the same
    """

    data_to_post = {
        "X": np.random.random(size=(data_size, len(sensors_str))).tolist(),
        "y": np.random.random(size=(data_size, len(sensors_str))).tolist(),
    }

    endpoint = f"{base_route}/anomaly/prediction"
    if resp_format is not None:
        endpoint += f"?format={resp_format}"

    resp = gordo_ml_server_client.post(endpoint, json=data_to_post)

    # From here, the response should be (pretty much) the same format from GET or POST
    assert resp.status_code == 200
    if resp_format in (None, "json"):
        assert "data" in resp.json
        data = server_utils.dataframe_from_dict(resp.json["data"])
    else:
        data = server_utils.dataframe_from_parquet_bytes(resp.data)

    # Only different between POST and GET is POST will return None for
    # start and end dates, because the server can't know what those are
    assert "start" in data
    assert "end" in data
    if data_to_post is not None:
        assert np.all(data["start"].isna())
        assert np.all(data["end"].isna())
    else:
        assert not np.any(data["start"].isna())
        assert not np.any(data["end"].isna())

    assert all(
        key in data
        for key in (
            "total-anomaly-scaled",
            "total-anomaly-unscaled",
            "tag-anomaly-scaled",
            "tag-anomaly-unscaled",
            "model-input",
            "model-output",
        )
    )


@pytest.mark.parametrize("resp_format", ("json", "parquet"))
def test_second_anomaly_prediction_endpoint(
    second_base_route,
    sensors_str,
    influxdb,
    gordo_ml_server_client,
    sensors,
    resp_format,
):
    data_to_post = {
        "X": np.random.random(size=(10, len(sensors_str))).tolist(),
        "y": np.random.random(size=(10, len(sensors_str))).tolist(),
    }

    endpoint = f"{second_base_route}/anomaly/prediction?format={resp_format}"

    resp = gordo_ml_server_client.post(endpoint, json=data_to_post)

    assert resp.status_code == 200
    if resp_format in (None, "json"):
        assert "data" in resp.json
        data = server_utils.dataframe_from_dict(resp.json["data"])
    else:
        data = server_utils.dataframe_from_parquet_bytes(resp.data)

    assert "smooth-tag-anomaly-scaled" not in data
    assert "smooth-tag-anomaly-unscaled" not in data
    assert "smooth-total-anomaly-scaled" not in data
    assert "smooth-total-anomaly-unscaled" not in data


@pytest.mark.parametrize("resp_format", ("json", "parquet"))
def test_second_anomaly_prediction_endpoint_all_columns(
    second_base_route,
    sensors_str,
    influxdb,
    gordo_ml_server_client,
    sensors,
    resp_format,
):
    data_to_post = {
        "X": np.random.random(size=(10, len(sensors_str))).tolist(),
        "y": np.random.random(size=(10, len(sensors_str))).tolist(),
    }

    endpoint = (
        f"{second_base_route}/anomaly/prediction?all_columns=yes&format={resp_format}"
    )

    resp = gordo_ml_server_client.post(endpoint, json=data_to_post)

    assert resp.status_code == 200
    if resp_format in (None, "json"):
        assert "data" in resp.json
        data = server_utils.dataframe_from_dict(resp.json["data"])
    else:
        data = server_utils.dataframe_from_parquet_bytes(resp.data)

    assert "smooth-tag-anomaly-scaled" in data
    assert "smooth-tag-anomaly-unscaled" in data
    assert "smooth-total-anomaly-scaled" in data
    assert "smooth-total-anomaly-unscaled" in data



"""Tests for gordo-client."""
# TODO: Move those tests to gordo-client project.

import json
import logging
import os
import string
from unittest import mock

import numpy as np
import pandas as pd
import pytest
import requests
from click.testing import CliRunner
from dateutil.parser import isoparse  # type: ignore
from gordo_client import Client, utils as client_utils
from gordo_client.forwarders import ForwardPredictionsIntoInflux
from gordo_client.io import (
    _handle_response,
    HttpUnprocessableEntity,
    BadGordoRequest,
    NotFound,
    ResourceGone,
)
from gordo_client.schemas import Machine as ClientMachine
from gordo_client.utils import PredictionResult
from gordo_dataset.data_provider import providers
from gordo_dataset.datasets import TimeSeriesDataset
from mock import patch, call
from sklearn.base import BaseEstimator

from gordo import serializer
from gordo_client.cli.client import gordo_client
from gordo.machine import Machine
from gordo.machine.model import utils as model_utils
from gordo.server import utils as server_utils


def test_client_get_metadata(gordo_project, ml_server):
    """
    Test client's ability to get metadata from some target
    """
    client = Client(project=gordo_project)

    metadata = client.get_metadata()
    assert isinstance(metadata, dict)

    # Can't get metadata for non-existent target
    assert client.get_metadata().get("no-such-target", None) is None


def test_client_get_dataset(gordo_project, metadata, ml_server):
    data_provider = providers.RandomDataProvider(min_size=10)
    client = Client(project=gordo_project, data_provider=data_provider)
    start = isoparse("2016-01-01T00:00:00+00:00")
    end = isoparse("2016-01-01T12:00:00+00:00")
    machine = Machine(**metadata)
    assert type(machine.dataset) is TimeSeriesDataset
    machine.dataset.row_filter_buffer_size = 12
    machine.dataset.n_samples_threshold = 10
    client_machine = ClientMachine(**machine.to_dict())
    dataset = client._get_dataset(client_machine, start, end)
    assert dataset.row_filter_buffer_size == 0
    assert dataset.n_samples_threshold == 0
    assert dataset.low_threshold is None
    assert dataset.high_threshold is None


def test_client_predict_specific_targets(gordo_project, gordo_single_target, ml_server):
    """
    Client.predict should filter any endpoints given to it.
    """
    client = Client(project=gordo_project)
    with mock.patch.object(
        Client,
        "predict_single_machine",
        return_value=PredictionResult("test-name", [], []),
    ) as patched:

        start = (isoparse("2016-01-01T00:00:00+00:00"),)
        end = isoparse("2016-01-01T12:00:00+00:00")

        # Should not call any predictions because this machine name doesn't exist
        with pytest.raises(NotFound):
            client.predict(start=start, end=end, targets=["non-existent-machine"])
            patched.assert_not_called()

        # Should be called once, for this machine.
        client.predict(start=start, end=end, targets=[gordo_single_target])
        patched.assert_called_once()


def test_client_download_model(gordo_project, gordo_single_target, ml_server):
    """
    Test client's ability to download the model
    """
    client = Client(project=gordo_project)

    models = client.download_model()
    assert isinstance(models, dict)
    assert isinstance(models[gordo_single_target], BaseEstimator)

    # Can't download model for non-existent target
    with pytest.raises(NotFound):
        client = Client(project=gordo_project)
        client.download_model(targets=["non-existent-target"])


@pytest.mark.parametrize("batch_size", (10, 100))
@pytest.mark.parametrize("use_parquet", (True, False))
def test_client_predictions_diff_batch_sizes(
    gordo_project,
    gordo_single_target,
    influxdb,
    influxdb_uri,
    influxdb_measurement,
    ml_server,
    batch_size: int,
    use_parquet: bool,
):
    """
    Run the prediction client with different batch-sizes and whether to use
    a data provider or not.
    """
    # Time range used in this test
    start, end = (
        isoparse("2016-01-01T00:00:00+00:00"),
        isoparse("2016-01-01T12:00:00+00:00"),
    )

    # Client only used within the this test
    test_client = client_utils.influx_client_from_uri(influxdb_uri)

    # Created measurements by prediction client with dest influx
    query = f"""
    SELECT *
    FROM "model-output"
    WHERE("machine" =~ /^{gordo_single_target}$/)
    """

    # Before predicting, influx destination db should be empty for 'predictions' measurement
    vals = test_client.query(query)
    assert len(vals) == 0

    data_provider = providers.InfluxDataProvider(
        measurement=influxdb_measurement,
        value_name="Value",
        client=client_utils.influx_client_from_uri(
            uri=influxdb_uri, dataframe_client=True
        ),
    )

    prediction_client = Client(
        project=gordo_project,
        data_provider=data_provider,
        prediction_forwarder=ForwardPredictionsIntoInflux(  # type: ignore
            destination_influx_uri=influxdb_uri
        ),
        batch_size=batch_size,
        use_parquet=use_parquet,
        parallelism=10,
    )

    assert len(prediction_client.get_machine_names()) == 2

    # Get predictions
    predictions = prediction_client.predict(start=start, end=end)
    assert isinstance(predictions, list)
    assert len(predictions) == 2

    name, predictions, error_messages = predictions[0]  # First dict of predictions
    assert isinstance(name, str)
    assert isinstance(predictions, pd.DataFrame)
    assert isinstance(error_messages, list)

    assert isinstance(predictions.index, pd.core.indexes.datetimes.DatetimeIndex)

    # This should have resulted in writting predictions to influx
    # Before predicting, influx destination db should be empty
    vals = test_client.query(query)
    assert (
        len(vals) > 0
    ), f"Expected new values in 'predictions' measurement, but found {vals}"


def test_client_metadata_revision(gordo_project, gordo_single_target, ml_server):
    prediction_client = Client(project=gordo_project)
    assert "revision" in prediction_client.get_available_machines()


@pytest.mark.parametrize(
    "args",
    [
        ["--help"],
        ["predict", "--help"],
        ["metadata", "--help"],
        ["download-model", "--help"],
    ],
)
def test_client_cli_basic(args):
    """
    Test that client specific subcommands exist
    """
    runner = CliRunner()
    out = runner.invoke(gordo_client, args=args)
    assert (
        out.exit_code == 0
    ), f"Expected output code 0 got '{out.exit_code}', {out.output}"


def test_client_cli_metadata(gordo_project, gordo_single_target, ml_server, tmpdir):
    """
    Test proper execution of client predict sub-command
    """
    runner = CliRunner()

    # Simple metadata fetching with single target
    out = runner.invoke(
        gordo_client,
        args=["--project", gordo_project, "metadata", "--target", gordo_single_target],
    )
    assert out.exit_code == 0
    assert gordo_single_target in out.output

    # Simple metadata fetching with all targets
    out = runner.invoke(gordo_client, args=["--project", gordo_project, "metadata"])
    assert out.exit_code == 0
    assert gordo_single_target in out.output

    # Save metadata to file
    output_file = os.path.join(tmpdir, "metadata.json")
    out = runner.invoke(
        gordo_client,
        args=[
            "--project",
            gordo_project,
            "metadata",
            "--output-file",
            output_file,
            "--target",
            gordo_single_target,
        ],
    )
    assert out.exit_code == 0, f"{out.exc_info}"
    assert os.path.exists(output_file)
    with open(output_file) as f:
        metadata = json.load(f)
        assert gordo_single_target in metadata


def test_client_cli_download_model(
    gordo_project, gordo_single_target, ml_server, tmpdir
):
    """
    Test proper execution of client predict sub-command
    """
    runner = CliRunner()

    # Empty output directory before downloading
    assert len(os.listdir(tmpdir)) == 0

    out = runner.invoke(
        gordo_client,
        args=[
            "--project",
            gordo_project,
            "download-model",
            str(tmpdir),
            "--target",
            gordo_single_target,
        ],
    )
    assert (
        out.exit_code == 0
    ), f"Expected output code 0 got '{out.exit_code}', {out.output}"

    # Output directory should not be empty any longer
    assert len(os.listdir(tmpdir)) > 0

    model_output_dir = os.path.join(tmpdir, gordo_single_target)
    assert os.path.isdir(model_output_dir)

    model = serializer.load(model_output_dir)
    assert isinstance(model, BaseEstimator)


@pytest.mark.parametrize("use_forwarder", [True, False])
@pytest.mark.parametrize("output_dir", [True, False])
@pytest.mark.parametrize("use_parquet", (True, False))
@pytest.mark.parametrize("session_config", ({}, {"headers": {}}))
def test_client_cli_predict(
    influxdb,
    influxdb_uri,
    gordo_project,
    gordo_single_target,
    ml_server,
    tmpdir,
    use_forwarder,
    trained_model_directory,
    output_dir,
    use_parquet,
    session_config,
):
    """
    Test ability for client to get predictions via CLI
    """
    runner = CliRunner()

    args = ["--metadata", "key,value", "--project", gordo_project]
    if session_config:
        args.extend(["--session-config", json.dumps(session_config)])

    args.extend(
        [
            "predict",
            "--parquet" if use_parquet else "--no-parquet",
            "2016-01-01T00:00:00Z",
            "2016-01-01T01:00:00Z",
        ]
    )

    influx_client = client_utils.influx_client_from_uri(
        uri=influxdb_uri, dataframe_client=True
    )
    query = """
        SELECT *
        FROM "resampled"
        """

    # Do we have forwarder args?
    if use_forwarder:
        args.extend(["--influx-uri", influxdb_uri, "--forward-resampled-sensors"])
        vals = influx_client.query(query)
        # There is no data there before we start doing things
        assert len(vals) == 0

    # Should it write out the predictions to dataframes in an output directory?
    if output_dir:
        args.extend(["--output-dir", str(tmpdir)])

    # Do we have a data provider, POST else GET requests
    args.extend(
        ["--data-provider", json.dumps(providers.RandomDataProvider().to_dict())]
    )

    # Run without any error
    with patch(
        "gordo_dataset.sensor_tag._asset_from_tag_name",
        side_effect=lambda *args, **kwargs: "default",
    ):
        out = runner.invoke(gordo_client, args=args)
    assert out.exit_code == 0, f"{out.output}"

    # If we activated forwarder and we had any actual data then there should
    # be resampled values in the influx
    if use_forwarder:
        vals = influx_client.query(query)
        assert len(vals) == 1
        assert len(vals["resampled"]) == 48
        influx_client.drop_measurement("resampled")

    # Did it save dataframes to output dir if specified?
    if output_dir:
        assert os.path.exists(os.path.join(tmpdir, f"{gordo_single_target}.csv.gz"))


@pytest.mark.parametrize(
    "should_fail,start_date,end_date",
    [
        (True, "1888-01-01T00:00:00Z", "1888-02-01T01:00:00Z"),  # Fail on bad dates
        (False, "2016-01-01T00:00:00Z", "2016-01-01T01:00:00Z"),  # pass on good dates
    ],
)
def test_client_cli_predict_non_zero_exit(
    should_fail,
    start_date,
    end_date,
    caplog,
    gordo_project,
    influxdb,
    influxdb_uri,
    influxdb_measurement,
    ml_server,
):
    """
    Test ability for client to get predictions via CLI
    """
    runner = CliRunner()

    # Should fail requesting dates which clearly don't exist.
    args = [
        "--metadata",
        "key,value",
        "--project",
        gordo_project,
        "predict",
        start_date,
        end_date,
    ]

    data_provider = providers.InfluxDataProvider(
        measurement=influxdb_measurement, value_name="Value", uri=influxdb_uri
    )

    args.extend(["--data-provider", json.dumps(data_provider.to_dict())])

    # Run without any error
    with caplog.at_level(logging.CRITICAL):
        with patch(
            "gordo_dataset.sensor_tag._asset_from_tag_name",
            side_effect=lambda *args, **kwargs: "default",
        ):
            out = runner.invoke(gordo_client, args=args)

    if should_fail:
        assert out.exit_code != 0, f"{out.output or out.exception}"
    else:
        assert out.exit_code == 0, f"{out.output or out.exception}"


@pytest.mark.parametrize("use_test_project_tags", [True, False])
def test_ml_server_dataframe_to_dict_and_back(sensors_str, use_test_project_tags):
    """
    Tests the flow of the server creating a dataframe from the model's data, putting into
    a dict of string to df. lists of values, and the client being able to reconstruct it back
    to the original dataframe (less the second level names)
    """
    # Run test with test project tag names
    if use_test_project_tags:
        tags = sensors_str
    # Run project with random names
    else:
        tags = [string.ascii_uppercase[i] for i in range(len(sensors_str))]

    # Some synthetic data
    original_input = np.random.random((10, len(tags)))
    model_output = np.random.random((10, len(tags)))

    # Convert this data into a dataframe with multi index columns
    df = model_utils.make_base_dataframe(tags, original_input, model_output)

    # Server then converts this into a dict which maps top level names to lists
    serialized = server_utils.dataframe_to_dict(df)

    # Client reproduces this dataframe
    df_clone = server_utils.dataframe_from_dict(serialized)

    # each subset of column under the top level names should be equal
    top_lvl_names = df.columns.get_level_values(0)
    for top_lvl_name in filter(lambda n: n not in ("start", "end"), top_lvl_names):
        assert np.allclose(df[top_lvl_name].values, df_clone[top_lvl_name].values)


def _machine(name: str) -> Machine:
    """
    Helper to build a basic Machine, only defining its name
    """
    from gordo_dataset.sensor_tag import SensorTag

    return Machine.from_config(
        config={
            "name": name,
            "dataset": {
                "tag_list": [SensorTag("tag-1", "foo"), SensorTag("tag-2", "foo")],
                "train_start_date": "2016-01-01T00:00:00Z",
                "train_end_date": "2016-01-05T00:00:00Z",
            },
            "model": {"sklearn.linear_model.LinearRegression": {}},
        },
        project_name="test-project",
    )


def test_exponential_sleep_time(caplog, gordo_project, ml_server):

    start, end = (
        isoparse("2016-01-01T00:00:00+00:00"),
        isoparse("2016-01-01T12:00:00+00:00"),
    )

    with caplog.at_level(logging.CRITICAL):
        with patch("gordo_client.client.sleep", return_value=None) as time_sleep:
            # We simulate repeating timeouts
            with patch("gordo_client.client._handle_response") as handle_response_mock:
                handle_response_mock.side_effect = TimeoutError()
                client = Client(project=gordo_project)

                client._send_prediction_request(
                    X=pd.DataFrame([123]),
                    y=None,
                    chunk=slice(0, 1),
                    machine=_machine("t1"),
                    start=start,
                    end=end,
                    revision="1234",
                )

                expected_calls = [call(8), call(16), call(32), call(64), call(128)]
                time_sleep.assert_has_calls(expected_calls)


def test__handle_response_errors():
    """
    Test expected error raising from gordo_client.io._handle_response
    """
    resp = requests.Response()
    resp.status_code = 422
    with pytest.raises(HttpUnprocessableEntity):
        _handle_response(resp)

    resp = requests.Response()
    resp.status_code = 403
    with pytest.raises(BadGordoRequest):
        _handle_response(resp)

    resp = requests.Response()
    resp.status_code = 502
    with pytest.raises(IOError):
        _handle_response(resp)


def test_client_set_revision_error(ml_server, gordo_project):
    """
    Client will raise an error if asking for a revision that doesn't exist
    """
    with pytest.raises(ResourceGone):
        client = Client(project=gordo_project)
        client.get_machine_names(revision="does-not-exist")



import json
import logging

import peewee
from playhouse.postgres_ext import (
    Model,
    PostgresqlExtDatabase,
    CharField,
    BinaryJSONField,
)
from playhouse.shortcuts import dict_to_model, model_to_dict

from .base import BaseReporter
from gordo.util.utils import capture_args
from gordo.machine import Machine as GordoMachine
from gordo.machine.machine import MachineEncoder
from .exceptions import ReporterException


logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


db = PostgresqlExtDatabase(None)


class PostgresReporterException(ReporterException):
    pass


class PostgresReporter(BaseReporter):
    """
    Reporter storing the :class:`gordo.machine.Machine` into a Postgres database.
    """

    db = db

    @capture_args
    def __init__(
        self,
        host: str,
        port: int = 5432,
        user: str = "postgres",
        password: str = "postgres",
        database: str = "postgres",
    ):
        self.host = host
        self.port = port
        self.user = user
        self.password = password
        self.database = database
        sql_parameters = {"host": self.host, "port": self.port, "user": self.user}
        if self.password:
            sql_parameters.update({"password": self.password})

        try:
            self.db.init(self.database, **sql_parameters)
            Machine.create_table(safe=True)
        except Exception as exc:
            raise PostgresReporterException(exc)

    def report(self, machine: GordoMachine):
        """
        Log a machine to Postgres where top level keys, 'name', 'dataset', 'model',
        and 'metadata' mappings to BinaryJSON fields.

        Parameters
        ----------
        machine: gordo.machine.Machine

        Returns
        -------
        None
        """
        try:
            with self.db.atomic():
                logger.info(f"Inserting machine {machine.name} in sql")  # type: ignore

                # Ensure it's serializable using MachineEncoder
                record = json.loads(json.dumps(machine.to_dict(), cls=MachineEncoder))
                model = dict_to_model(Machine, record, ignore_unknown=True)
                try:
                    Machine.get(Machine.name == machine.name)
                except peewee.DoesNotExist:
                    model.save()
                else:
                    query = Machine.update(**model_to_dict(model)).where(
                        Machine.name == machine.name
                    )
                    query.execute()

        except Exception as exc:
            raise PostgresReporterException(exc)


class Machine(Model):
    name = CharField(index=True, unique=True)
    dataset = BinaryJSONField()
    model = BinaryJSONField()
    metadata = BinaryJSONField()

    class Meta:
        primary_key = False
        database = db
        table_name = "machine"

    def __repr__(self):
        return f"Machine {self.__data__} "
